""" builder.py: ModelBuilder class builds and updates the unit commitment problem.
"""

from .input import SystemInput
import logging

from gurobipy import GRB
import gurobipy as gp
import pandas as pd

import pownet.modeling as modeling
from pownet.modeling import PowerSystemModel
from pownet.data_utils import get_unit_hour_from_varnam, get_edge_hour_from_varname

logger = logging.getLogger(__name__)


class ModelBuilder:
    """Build an instance of the Unit Commitment Problem.
    We keep the instance of the Gurobi model throughout simulation to minimize
    the overhead of creating a new model for each iteration and also benefit from
    the warm start feature of Gurobi. This means the following:
    * Variable names are unchanged across iterations
    * Only a subset of constraints are updated at each iteration
      - Remove and build the objective function
      - Update the upper bounds of variables
      - Remove and add constraints that are time-dependent, such as

    Define the lower/upper bounds explicitly at the variable level instead of
    defining them as constraints. This approach can leverage gurobi's presolve
    and reduce the size of the model.

    """

    def __init__(self, inputs: SystemInput) -> None:
        self.inputs = inputs
        self.timesteps = range(1, self.inputs.sim_horizon + 1)
        self.model: gp.Model = None

        # Variables
        # Thermal units
        self.pthermal = gp.tupledict()
        self.status = gp.tupledict()
        self.startup = gp.tupledict()
        self.shutdown = gp.tupledict()
        self.vpower = gp.tupledict()
        self.vpowerbar = gp.tupledict()
        self.spin = gp.tupledict()
        self.pthermal_curtail = gp.tupledict()  # for must-take units

        # Energy storage
        self.pcharge = gp.tupledict()
        self.pdischarge = gp.tupledict()

        self.charge_state = gp.tupledict()  # State of charge
        self.ucharge = gp.tupledict()  # Charging indicator
        self.udischarge = gp.tupledict()  # Discharging indicator

        # Renewable energy and import sources
        self.phydro = gp.tupledict()
        self.phydro_curtail = gp.tupledict()
        self.uhydro = gp.tupledict()

        self.psolar = gp.tupledict()
        self.psolar_curtail = gp.tupledict()
        self.usolar = gp.tupledict()

        self.pwind = gp.tupledict()
        self.pwind_curtail = gp.tupledict()
        self.uwind = gp.tupledict()

        self.pimp = gp.tupledict()  # import
        self.pimp_curtail = gp.tupledict()
        self.uimp = gp.tupledict()

        # Node variables
        self.pos_pmismatch = gp.tupledict()
        self.neg_pmismatch = gp.tupledict()

        # Energy-flow variables
        self.flow = gp.tupledict()
        self.theta = gp.tupledict()

        # System-wide variables
        self.spin_shortfall = gp.tupledict()

        # Time-independent terms in the objective function
        self.thermal_fixed_expr = gp.LinExpr()
        self.thermal_opex_expr = gp.LinExpr()
        self.thermal_startup_expr = gp.LinExpr()
        self.thermal_curtail_expr = gp.LinExpr()

        self.load_shortfall_penalty_expr = gp.LinExpr()
        self.spin_shortfall_penalty_expr = gp.LinExpr()

        # Thermal unit constraints
        self.c_link_uvw_init = gp.tupledict()
        self.c_link_uvw = gp.tupledict()
        self.c_link_pthermal = gp.tupledict()
        self.c_link_pu_lower = gp.tupledict()
        self.c_link_pu_upper = gp.tupledict()
        self.c_min_down_init = gp.tupledict()
        self.c_min_up_init = gp.tupledict()
        self.c_min_down = gp.tupledict()
        self.c_min_up = gp.tupledict()
        self.c_peak_down_bound = gp.tupledict()
        self.c_peak_up_bound = gp.tupledict()
        self.c_ramp_down_init = gp.tupledict()
        self.c_ramp_up_init = gp.tupledict()
        self.c_ramp_down = gp.tupledict()
        self.c_ramp_up = gp.tupledict()
        self.c_link_spin = gp.tupledict()
        self.c_link_ppbar = gp.tupledict()
        self.c_thermal_curtail = gp.tupledict()

        # System-wide constraints
        self.c_ref_node = gp.tupledict()
        self.c_angle_diff = gp.tupledict()
        self.c_kirchhoff = gp.tupledict()
        self.c_flow_balance = gp.tupledict()
        self.c_reserve_req = gp.tupledict()

        self.c_hydro_curtail_ess = gp.tupledict()
        self.c_daily_hydro_curtail_ess = gp.tupledict()
        self.c_solar_curtail_ess = gp.tupledict()
        self.c_wind_curtail_ess = gp.tupledict()
        self.c_import_curtail_ess = gp.tupledict()

        # These constraints act as the upper bound
        # and provides indicators if the unit is online
        self.c_link_hydro_pu = gp.tupledict()
        self.c_link_daily_hydro_pu = gp.tupledict()
        self.c_link_solar_pu = gp.tupledict()
        self.c_link_wind_pu = gp.tupledict()
        self.c_link_import_pu = gp.tupledict()

        self.c_hydro_limit_daily = gp.tupledict()

        # Energy storage constraints
        self.c_link_ess_charge = gp.tupledict()
        self.c_link_discharge = gp.tupledict()
        self.c_link_ess_state = gp.tupledict()
        self.c_unit_ess_balance_init = gp.tupledict()
        self.c_unit_ess_balance = gp.tupledict()

    def add_variables(self, step_k: int) -> None:
        """Add variables to the model.

        Thermal-unit variables:
        -----------------
        - `pthermal`: Power output by a thermal unit (also called dispatch). Unit: MW.
        - `vpower`: Power output *ABOVE* the minimum capacity of a thermal unit. Unit: MW.
        - `vpowerbar`: Maximum power output *ABOVE* the minimum capacity of a thermal unit. Unit: MW.
        - `status`: Indicator of online status. On = 1 and off = 0. Unitless.
        - `startup`: Indicator if a unit is shutting down. Unitless.
        - `shutdown`: Indicator if a unit is starting up. Unitless.
        - `spin` (optional): Spinning reserve. Unit: MW.
        - `pthermal_curtail`: Curtailed power output by a thermal unit. Unit: MW.

        Energy-storage variables:
        -----------------
        - `pcharge`: Power charging an energy storage system. Unit: MW.
        - `pdischarge`: Power discharging an energy storage system. Unit: MW.
        - `charge_state`: State of charge of an energy storage system. Unit: MWh.
        - `ucharge`: Indicator that an ESS is charging. Unitless.
        - `udischarge`: Indicator that an ESS is discharging. Unitless.

        Renewable energy and import variables:
        -----------------
        - `phydro`: Hydropower output. Unit: MW.
        - `psolar`: Solar power output. Unit: MW.
        - `pwind`: Wind power output. Unit: MW.
        - `pimp`: Import power output. Unit: MW.
        - `phydro_curtail`: Curtailed hydropower output. Unit: MW.
        - `psolar_curtail`: Curtailed solar power output. Unit: MW.
        - `pwind_curtail`: Curtailed wind power output. Unit: MW.
        - `pimp_curtail`: Curtailed import power output. Unit: MW.

        Node variables:
        -----------------
        - `pos_pmismatch`: Positive power mismatch. Unit: MW.
        - `neg_pmismatch`: Negative power mismatch. Unit: MW.
        - `theta`: Voltage angle. Unit: Radians.

        Flow variables:
        -----------------
        - `flow`: Power flow on transmission lines. Unit: MW/hr.

        System variables:
        -----------------
        - `spin_shortfall`: Spinning reserve shortfall. Unit: MW.

        Args:
            step_k: The current step of the simulation.

        Returns:
            None

        """
        ################################
        # Variables with fixed upper bounds
        ################################
        var_with_u_tuples = [
            ("phydro", self.inputs.hydro_units, self.inputs.hydro_contracted_capacity),
            (
                "phydro_curtail",
                self.inputs.hydro_units,
                self.inputs.hydro_contracted_capacity,
            ),
            ("psolar", self.inputs.solar_units, self.inputs.solar_contracted_capacity),
            (
                "psolar_curtail",
                self.inputs.solar_units,
                self.inputs.solar_contracted_capacity,
            ),
            ("pwind", self.inputs.wind_units, self.inputs.wind_contracted_capacity),
            (
                "pwind_curtail",
                self.inputs.wind_units,
                self.inputs.wind_contracted_capacity,
            ),
            ("pimp", self.inputs.import_units, self.inputs.import_contracted_capacity),
            (
                "pimp_curtail",
                self.inputs.import_units,
                self.inputs.import_contracted_capacity,
            ),
            ("pcharge", self.inputs.storage_units, self.inputs.ess_max_charge),
            ("pdischarge", self.inputs.storage_units, self.inputs.ess_max_discharge),
        ]

        for varname, units, capacity_dict in var_with_u_tuples:
            setattr(
                self,
                f"{varname}",
                self.model.addVars(
                    units,
                    self.timesteps,
                    lb=0,
                    ub={
                        (unit, t): capacity_dict[unit]
                        for t in self.timesteps
                        for unit in units
                    },
                    name=varname,
                ),
            )

        ################################
        # Variables with variable upper bounds
        ################################

        var_with_variable_ub_tuples = [
            ("pthermal", "thermal_units", self.inputs.thermal_derated_capacity),
            ("vpower", "thermal_units", self.inputs.thermal_derated_capacity),
            ["vpowerbar", "thermal_units", self.inputs.thermal_derated_capacity],
            (
                "pthermal_curtail",
                "thermal_must_take_units",
                self.inputs.thermal_derated_capacity,
            ),
            ("charge_state", "storage_units", self.inputs.ess_derated_capacity),
        ]

        for varname, unit_type, capacity_df in var_with_variable_ub_tuples:
            # Update the self attribute directly
            setattr(
                self,
                varname,
                modeling.add_var_with_variable_ub(
                    model=self.model,
                    varname=varname,
                    timesteps=self.timesteps,
                    step_k=step_k,
                    units=getattr(self.inputs, unit_type),
                    capacity_df=capacity_df,
                ),
            )

        ################################
        # Variables without upper bounds
        ################################

        vars_tuples = [
            ("pos_pmismatch", "demand_nodes"),
            ("neg_pmismatch", "demand_nodes"),
        ]

        for varname, unit_type in vars_tuples:
            setattr(
                self,
                varname,
                self.model.addVars(
                    getattr(self.inputs, unit_type),
                    self.timesteps,
                    name=varname,
                ),
            )

        # Spinning reserve can be modeled with or without
        # the spin variable
        if self.inputs.use_spin_var:
            self.spin = self.model.addVars(
                self.inputs.thermal_units,
                self.timesteps,
                name="spin",
            )

        # System-wide spinning reserve shortfall
        self.spin_shortfall = self.model.addVars(
            self.timesteps,
            name="spin_shortfall",
        )

        ################################
        # Binary variables
        ################################
        var_binary_tuples = [
            ("status", "thermal_units"),
            ("startup", "thermal_units"),
            ("shutdown", "thermal_units"),
            ("ucharge", "storage_units"),
            ("udischarge", "storage_units"),
            ("uhydro", "hydro_units"),
            ("usolar", "solar_units"),
            ("uwind", "wind_units"),
            ("uimp", "import_units"),
        ]
        for varname, unit_type in var_binary_tuples:
            setattr(
                self,
                varname,
                self.model.addVars(
                    getattr(self.inputs, unit_type),
                    self.timesteps,
                    vtype=GRB.BINARY,
                    name=varname,
                ),
            )

        ################################
        # Flow variables
        ################################

        self.flow = modeling.add_var_flow(
            model=self.model,
            timesteps=self.timesteps,
            step_k=step_k,
            edges=self.inputs.edges,
            line_capacity_factor=self.inputs.line_capacity_factor,
            line_capacity=self.inputs.line_capacity,
        )

        if self.inputs.dc_opf == "voltage_angle":
            self.theta = modeling.add_var_voltage_angle(
                model=self.model,
                timesteps=self.timesteps,
                nodes=self.inputs.nodes,
            )

        ################################
        # End of adding variables
        ################################

        self.model.update()

    def _build_rnw_import_storage_objfunc_terms(self, step_k: int) -> gp.LinExpr:
        """Build the objective function for renewable energy and import sources."""
        variable_unit_contracts_tuple = [
            (self.phydro, self.inputs.hydro_units, self.inputs.nondispatch_contracts),
            # hourly hydro
            (
                self.phydro_curtail,
                self.inputs.hydro_must_take_units,
                self.inputs.nondispatch_contracts,
            ),
            # Daily hydro
            (
                self.phydro_curtail,
                self.inputs.daily_hydro_must_take_units,
                self.inputs.nondispatch_contracts,
            ),
            (self.psolar, self.inputs.solar_units, self.inputs.nondispatch_contracts),
            (
                self.psolar_curtail,
                self.inputs.solar_must_take_units,
                self.inputs.nondispatch_contracts,
            ),
            (self.pwind, self.inputs.wind_units, self.inputs.nondispatch_contracts),
            (
                self.pwind_curtail,
                self.inputs.wind_must_take_units,
                self.inputs.nondispatch_contracts,
            ),
            (self.pimp, self.inputs.import_units, self.inputs.nondispatch_contracts),
            (
                self.pimp_curtail,
                self.inputs.import_must_take_units,
                self.inputs.nondispatch_contracts,
            ),
            (self.pdischarge, self.inputs.storage_units, self.inputs.ess_contracts),
        ]

        rnw_import_storage_expr = gp.LinExpr()
        for var_dict, units, contracts in variable_unit_contracts_tuple:
            cost_coeffs = modeling.get_marginal_cost_coeff(
                step_k=step_k,
                timesteps=self.timesteps,
                units=units,
                nondispatch_contracts=contracts,
                contract_costs=self.inputs.contract_costs,
            )
            rnw_import_storage_expr.add(var_dict.prod(cost_coeffs))
        return rnw_import_storage_expr

    def set_objfunc(self, step_k: int) -> None:
        """The objective function has four components: unit fixed cost, unit variable cost,
        unit start-up cost, and shortfall cost. THe objective function is to minimize the
        total cost of generation and shortfall. Some terms are time-dependent, such as
        the unit variable cost.

        Args:
            step_k: The current step of the simulation.

        Returns:
            None
        """

        ################################
        # Thermal units
        ################################
        thermal_fixed_coeffs = modeling.get_thermal_fixed_coeff(
            timesteps=self.timesteps,
            thermal_units=self.inputs.thermal_units,
            thermal_fixed_cost=self.inputs.thermal_fixed_cost,
            thermal_rated_capacity=self.inputs.thermal_rated_capacity,
        )
        self.thermal_fixed_expr = self.status.prod(thermal_fixed_coeffs)

        thermal_opex_coeffs = modeling.get_thermal_opex_coeff(
            step_k=step_k,
            timesteps=self.timesteps,
            thermal_units=self.inputs.thermal_units,
            thermal_opex=self.inputs.thermal_opex,
            fuel_contracts=self.inputs.fuel_contracts,
            contract_costs=self.inputs.contract_costs,
            thermal_heat_rate=self.inputs.thermal_heat_rate,
        )

        self.thermal_opex_expr = self.pthermal.prod(thermal_opex_coeffs)

        thermal_startup_coeffs = modeling.get_thermal_startup_coeff(
            timesteps=self.timesteps,
            thermal_units=self.inputs.thermal_units,
            thermal_startup_cost=self.inputs.thermal_startup_cost,
            thermal_rated_capacity=self.inputs.thermal_rated_capacity,
        )
        self.thermal_startup_expr = self.startup.prod(thermal_startup_coeffs)

        # Curtailment cost is the same as the OPEX cost
        self.thermal_curtail_expr = self.pthermal_curtail.prod(thermal_opex_coeffs)

        ################################
        # Shortfall penalties
        ################################

        # The cost of shortfall is the slack variable (pos_pmismatch) needed to meet demand
        self.load_shortfall_penalty_expr = self.inputs.load_shortfall_penalty_factor * (
            gp.quicksum(self.pos_pmismatch)
        ) + (self.inputs.load_curtail_penalty_factor * gp.quicksum(self.neg_pmismatch))

        # Penalize the system if it cannot meet the spinning reserve requirement
        self.spin_shortfall_penalty_expr = (
            self.inputs.spin_shortfall_penalty_factor * gp.quicksum(self.spin_shortfall)
        )

        ################################
        # Renewable energy, import sources, and energy storage
        ################################
        rnw_import_expr = self._build_rnw_import_storage_objfunc_terms(step_k=step_k)

        ################################
        # Set the objective function
        ################################

        self.model.setObjective(
            (
                self.thermal_fixed_expr
                + self.thermal_opex_expr
                + self.thermal_startup_expr
                + self.thermal_curtail_expr
                + self.load_shortfall_penalty_expr
                + self.spin_shortfall_penalty_expr
                + rnw_import_expr
            ),
            sense=GRB.MINIMIZE,
        )

        self.model.update()

    def _add_hourly_hydropower_constraints(self, step_k: int) -> None:
        self.c_link_hydro_pu = modeling.add_c_link_unit_pu(
            model=self.model,
            pdispatch=self.phydro,
            u=self.uhydro,
            type="hydro",
            timesteps=self.timesteps,
            step_k=step_k,
            units=self.inputs.hydro_unit_node.keys(),
            capacity_df=self.inputs.hydro_capacity,
        )
        self.c_hydro_curtail_ess = modeling.add_c_unit_curtail_ess(
            model=self.model,
            pdispatch=self.phydro,
            pcurtail=self.phydro_curtail,
            pcharge=self.pcharge,
            type="hydro",
            timesteps=self.timesteps,
            step_k=step_k,
            units=self.inputs.hydro_unit_node.keys(),
            capacity_df=self.inputs.hydro_capacity,
            ess_attached=self.inputs.ess_hydro_units,
        )

    def _add_daily_hydropower_constraints(self, step_k: int) -> None:
        # (1) Define the daily upper bound.
        self.c_hydro_limit_daily = modeling.add_c_hydro_limit_daily(
            model=self.model,
            phydro=self.phydro,
            step_k=step_k,
            sim_horizon=self.inputs.sim_horizon,
            hydro_units=self.inputs.daily_hydro_unit_node.keys(),
            hydro_capacity=self.inputs.daily_hydro_capacity,
        )
        # (2) The curtailment is enforced for the day and not the hour.
        self.c_daily_hydro_curtail_ess = modeling.add_c_unit_curtail_ess_daily(
            model=self.model,
            pdispatch=self.phydro,
            pcurtail=self.phydro_curtail,
            pcharge=self.pcharge,
            type="hydro",
            sim_horizon=self.inputs.sim_horizon,
            step_k=step_k,
            units=self.inputs.daily_hydro_unit_node.keys(),
            capacity_df=self.inputs.daily_hydro_capacity,
            ess_attached=self.inputs.ess_daily_hydro_units,
        )

    def _add_hydropower_constraints(self, step_k: int) -> None:

        self._add_hourly_hydropower_constraints(step_k=step_k)
        self._add_daily_hydropower_constraints(step_k=step_k)

        # Does not update every step_k for this constraint
        self.c_link_daily_hydro_pu = modeling.add_c_link_unit_pu_constant(
            model=self.model,
            pdispatch=self.phydro,
            u=self.uhydro,
            timesteps=self.timesteps,
            units=self.inputs.daily_hydro_unit_node.keys(),
            contracted_capacity=self.inputs.hydro_contracted_capacity,
        )

    def _update_hydropower_constraints(self, step_k: int) -> None:
        # Hourly constraints
        self.model.remove(self.c_link_hydro_pu)
        self.model.remove(self.c_hydro_curtail_ess)
        self._add_hourly_hydropower_constraints(step_k=step_k)

        self.model.remove(self.c_hydro_limit_daily)
        self.model.remove(self.c_daily_hydro_curtail_ess)
        self._add_daily_hydropower_constraints(step_k=step_k)

    def _add_unit_link_pu(self, step_k: int) -> None:
        # Define parameters
        unit_params = {
            "solar": {
                "p": self.psolar,
                "u": self.usolar,
                "units": self.inputs.solar_units,
                "capacity_df": self.inputs.solar_capacity,
            },
            "wind": {
                "p": self.pwind,
                "u": self.uwind,
                "units": self.inputs.wind_units,
                "capacity_df": self.inputs.wind_capacity,
            },
            "import": {
                "p": self.pimp,
                "u": self.uimp,
                "units": self.inputs.import_units,
                "capacity_df": self.inputs.import_capacity,
            },
        }
        for type, params in unit_params.items():
            setattr(
                self,
                f"c_link_{type}_pu",
                modeling.add_c_link_unit_pu(
                    model=self.model,
                    pdispatch=params["p"],
                    u=params["u"],
                    type=type,
                    timesteps=self.timesteps,
                    step_k=step_k,
                    units=params["units"],
                    capacity_df=params["capacity_df"],
                ),
            )

    def _add_curtail_ess_constraints(self, step_k: int) -> None:
        # Not implementing ESS for thermal units
        unit_types_with_ess = {
            "solar": {
                "pdispatch": self.psolar,
                "pcurtail": self.psolar_curtail,
                "pcharge": self.pcharge,
                "units": self.inputs.solar_units,
                "capacity_df": self.inputs.solar_capacity,
                "ess_attached": self.inputs.ess_solar_units,
            },
            "wind": {
                "pdispatch": self.pwind,
                "pcurtail": self.pwind_curtail,
                "pcharge": self.pcharge,
                "units": self.inputs.wind_units,
                "capacity_df": self.inputs.wind_capacity,
                "ess_attached": self.inputs.ess_wind_units,
            },
            "import": {
                "pdispatch": self.pimp,
                "pcurtail": self.pimp_curtail,
                "pcharge": self.pcharge,
                "units": self.inputs.import_units,
                "capacity_df": self.inputs.import_capacity,
                "ess_attached": {},  # No ESS attached to import sources
            },
        }

        for unit_type, params in unit_types_with_ess.items():
            setattr(
                self,
                f"c_{unit_type}_curtail_ess",
                modeling.add_c_unit_curtail_ess(
                    model=self.model,
                    pdispatch=params["pdispatch"],
                    pcurtail=params["pcurtail"],
                    pcharge=params["pcharge"],
                    type=unit_type,
                    timesteps=self.timesteps,
                    step_k=step_k,
                    units=params["units"],
                    capacity_df=params["capacity_df"],
                    ess_attached=params["ess_attached"],
                ),
            )

    def add_constraints(self, step_k: int, init_conds: dict) -> None:
        ################################
        # Therma-unit constraints
        ################################

        self.c_link_uvw_init = modeling.add_c_link_uvw_init(
            model=self.model,
            u=self.status,
            v=self.startup,
            w=self.shutdown,
            initial_u=init_conds["initial_u"],
            thermal_units=self.inputs.thermal_units,
        )
        self.c_link_uvw = modeling.add_c_link_uvw(
            model=self.model,
            u=self.status,
            v=self.startup,
            w=self.shutdown,
            sim_horizon=self.inputs.sim_horizon,
            thermal_units=self.inputs.thermal_units,
        )
        self.c_link_pthermal = modeling.add_c_link_pthermal(
            model=self.model,
            pthermal=self.pthermal,
            p=self.vpower,
            u=self.status,
            timesteps=self.timesteps,
            thermal_units=self.inputs.thermal_units,
            thermal_min_capacity=self.inputs.thermal_min_capacity,
        )
        self.c_link_pu_lower = modeling.add_c_link_pu_lower(
            model=self.model,
            pthermal=self.pthermal,
            u=self.status,
            timesteps=self.timesteps,
            thermal_units=self.inputs.thermal_units,
            thermal_min_capacity=self.inputs.thermal_min_capacity,
        )
        self.c_link_pu_upper = modeling.add_c_link_pu_upper(
            model=self.model,
            pbar=self.vpowerbar,
            u=self.status,
            timesteps=self.timesteps,
            step_k=step_k,
            thermal_units=self.inputs.thermal_units,
            thermal_min_capacity=self.inputs.thermal_min_capacity,
            thermal_derated_capacity=self.inputs.thermal_derated_capacity,
        )
        self.c_min_down_init = modeling.add_c_min_down_init(
            model=self.model,
            u=self.status,
            sim_horizon=self.inputs.sim_horizon,
            thermal_units=self.inputs.thermal_units,
            initial_min_off=init_conds["initial_min_off"],
        )
        self.c_min_up_init = modeling.add_c_min_up_init(
            model=self.model,
            u=self.status,
            sim_horizon=self.inputs.sim_horizon,
            thermal_units=self.inputs.thermal_units,
            initial_min_on=init_conds["initial_min_on"],
        )
        self.c_min_down = modeling.add_c_min_down(
            model=self.model,
            u=self.status,
            w=self.shutdown,
            sim_horizon=self.inputs.sim_horizon,
            thermal_units=self.inputs.thermal_units,
            TD=self.inputs.TD,
        )
        self.c_min_up = modeling.add_c_min_up(
            model=self.model,
            u=self.status,
            v=self.startup,
            sim_horizon=self.inputs.sim_horizon,
            thermal_units=self.inputs.thermal_units,
            TU=self.inputs.TU,
        )
        self.c_thermal_curtail = modeling.add_c_thermal_curtail(
            model=self.model,
            pthermal=self.pthermal,
            pthermal_curtail=self.pthermal_curtail,
            pcharge=self.pcharge,
            timesteps=self.timesteps,
            step_k=step_k,
            thermal_derated_capacity=self.inputs.thermal_derated_capacity,
            thermal_must_take_units=self.inputs.thermal_must_take_units,
            ess_attached=self.inputs.ess_thermal_units,
        )

        # Currently not implemented because we set SD = SU = ramping
        # self.c_peak_down_bound = modeling.add_c_peak_down_bound(
        #     model=self.model,
        #     p=self.vpower,
        #     spin=self.spin,
        #     u=self.status,
        #     v=self.startup,
        #     w=self.shutdown,
        #     sim_horizon=self.inputs.sim_horizon,
        #     step_k=step_k,
        #     thermal_units=self.inputs.thermal_units,
        #     thermal_min_capacity=self.inputs.thermal_min_capacity,
        #     thermal_derated_capacity=self.inputs.thermal_derated_capacity,
        #     SD=self.inputs.SD,
        #     SU=self.inputs.SU,
        #     TU=self.inputs.TU,
        # )

        # self.c_peak_up_bound = modeling.add_c_peak_up_bound(
        #     model=self.model,
        #     p=self.vpower,
        #     spin=self.spin,
        #     u=self.status,
        #     v=self.startup,
        #     w=self.shutdown,
        #     sim_horizon=self.inputs.sim_horizon,
        #     step_k=step_k,
        #     thermal_units=self.inputs.thermal_units,
        #     thermal_min_capacity=self.inputs.thermal_min_capacity,
        #     thermal_derated_capacity=self.inputs.thermal_derated_capacity,
        #     SD=self.inputs.SD,
        #     SU=self.inputs.SU,
        #     TU=self.inputs.TU,
        # )

        self.c_ramp_down_init = modeling.add_c_ramp_down_init(
            model=self.model,
            p=self.vpower,
            w=self.shutdown,
            thermal_units=self.inputs.thermal_units,
            initial_p=init_conds["initial_p"],
            initial_u=init_conds["initial_u"],
            thermal_min_capacity=self.inputs.thermal_min_capacity,
            RD=self.inputs.RD,
            SD=self.inputs.SD,
        )
        self.c_ramp_up_init = modeling.add_c_ramp_up_init(
            model=self.model,
            pbar=self.vpowerbar,
            u=self.status,
            v=self.startup,
            thermal_units=self.inputs.thermal_units,
            initial_p=init_conds["initial_p"],
            thermal_min_capacity=self.inputs.thermal_min_capacity,
            RU=self.inputs.RU,
            SU=self.inputs.SU,
        )
        self.c_ramp_down = modeling.add_c_ramp_down(
            model=self.model,
            p=self.vpower,
            u=self.status,
            w=self.shutdown,
            sim_horizon=self.inputs.sim_horizon,
            thermal_units=self.inputs.thermal_units,
            thermal_min_capacity=self.inputs.thermal_min_capacity,
            RD=self.inputs.RD,
            SD=self.inputs.SD,
        )
        self.c_ramp_up = modeling.add_c_ramp_up(
            model=self.model,
            p=self.vpower,
            pbar=self.vpowerbar,
            u=self.status,
            v=self.startup,
            sim_horizon=self.inputs.sim_horizon,
            thermal_units=self.inputs.thermal_units,
            thermal_min_capacity=self.inputs.thermal_min_capacity,
            RU=self.inputs.RU,
            SU=self.inputs.SU,
        )

        ################################
        # Energy-storage constraints
        ################################

        self.c_link_ess_charge = modeling.add_c_link_ess_charge(
            model=self.model,
            pcharge=self.pcharge,
            ucharge=self.ucharge,
            timesteps=self.timesteps,
            units=self.inputs.storage_units,
            max_charge=self.inputs.ess_max_charge,
        )

        self.c_link_ess_dischage = modeling.add_c_link_ess_discharge(
            model=self.model,
            pdischarge=self.pdischarge,
            udischarge=self.udischarge,
            timesteps=self.timesteps,
            units=self.inputs.storage_units,
            max_discharge=self.inputs.ess_max_discharge,
        )

        self.c_link_ess_state = modeling.add_c_link_ess_state(
            model=self.model,
            ucharge=self.ucharge,
            udischarge=self.udischarge,
            timesteps=self.timesteps,
            units=self.inputs.storage_units,
        )

        self.c_unit_ess_balance_init = modeling.add_c_unit_ess_balance_init(
            model=self.model,
            pcharge=self.pcharge,
            pdischarge=self.pdischarge,
            charge_state=self.charge_state,
            units=self.inputs.storage_units,
            charge_state_init=init_conds["initial_charge_state"],
            charge_efficiency=self.inputs.ess_charge_efficiency,
            discharge_efficiency=self.inputs.ess_discharge_efficiency,
            self_discharge_rate=self.inputs.ess_self_discharge_rate,
        )

        self.c_unit_ess_balance = modeling.add_c_unit_ess_balance(
            model=self.model,
            pcharge=self.pcharge,
            pdischarge=self.pdischarge,
            charge_state=self.charge_state,
            units=self.inputs.storage_units,
            sim_horizon=self.inputs.sim_horizon,
            charge_efficiency=self.inputs.ess_charge_efficiency,
            discharge_efficiency=self.inputs.ess_discharge_efficiency,
            self_discharge_rate=self.inputs.ess_self_discharge_rate,
        )

        ################################
        # Power flow constraints
        ################################

        if self.inputs.dc_opf == "voltage_angle":
            self.c_ref_node = modeling.add_c_ref_node(
                model=self.model,
                theta=self.theta,
                timesteps=self.timesteps,
                max_demand_node=self.inputs.max_demand_node,
            )
            self.c_angle_diff = modeling.add_c_angle_diff(
                model=self.model,
                flow=self.flow,
                theta=self.theta,
                timesteps=self.timesteps,
                step_k=step_k,
                edges=self.inputs.edges,
                susceptance=self.inputs.susceptance,
            )
        elif self.inputs.dc_opf == "kirchhoff":
            self.c_kirchhoff = modeling.add_c_kirchhoff(
                model=self.model,
                flow=self.flow,
                timesteps=self.timesteps,
                step_k=step_k,
                edges=self.inputs.edges,
                cycle_map=self.inputs.cycle_map,
                susceptance=self.inputs.susceptance,
            )
        else:
            raise ValueError(f"Invalid DC-OPF parameter: {self.dc_opf}.")

        self.c_flow_balance = modeling.add_c_flow_balance(
            model=self.model,
            pthermal=self.pthermal,
            phydro=self.phydro,
            psolar=self.psolar,
            pwind=self.pwind,
            pimp=self.pimp,
            pcharge=self.pcharge,
            pdis=self.pdischarge,
            pos_pmismatch=self.pos_pmismatch,
            neg_pmismatch=self.neg_pmismatch,
            flow=self.flow,
            timesteps=self.timesteps,
            step_k=step_k,
            thermal_units=self.inputs.thermal_units,
            hydro_units=self.inputs.hydro_units,
            solar_units=self.inputs.solar_units,
            wind_units=self.inputs.wind_units,
            import_units=self.inputs.import_units,
            nodes=self.inputs.nodes,
            node_edge=self.inputs.node_edge,
            node_generator=self.inputs.node_generator,
            charge_storage=self.inputs.ess_substation_units,
            discharge_storage=self.inputs.ess_attach_unit,
            demand_nodes=self.inputs.demand_nodes,
            demand=self.inputs.demand,
            line_loss_factor=self.inputs.line_loss_factor,
        )

        if self.inputs.use_spin_var:
            self.c_link_spin = modeling.add_c_link_spin(
                model=self.model,
                p=self.vpower,
                pbar=self.vpowerbar,
                spin=self.spin,
                timesteps=self.timesteps,
                thermal_units=self.inputs.thermal_units,
            )
            self.c_reserve_req = modeling.add_c_reserve_req_1(
                model=self.model,
                spin=self.spin,
                spin_shortfall=self.spin_shortfall,
                timesteps=self.timesteps,
                step_k=step_k,
                thermal_units=self.inputs.thermal_units,
                spin_requirement=self.inputs.spin_requirement,
            )
        else:
            self.c_link_ppbar = modeling.add_c_link_ppbar(
                model=self.model,
                p=self.vpower,
                pbar=self.vpowerbar,
                timesteps=self.timesteps,
                thermal_units=self.inputs.thermal_units,
            )
            self.c_reserve_req = modeling.add_c_reserve_req_2(
                model=self.model,
                pbar=self.vpowerbar,
                u=self.status,
                spin_shortfall=self.spin_shortfall,
                timesteps=self.timesteps,
                step_k=step_k,
                thermal_units=self.inputs.thermal_units,
                thermal_min_capacity=self.inputs.thermal_min_capacity,
                demand_nodes=self.inputs.demand_nodes,
                demand=self.inputs.demand,
                spin_requirement=self.inputs.spin_requirement,
            )

        ################################
        # Non-dispatchables
        ################################

        # Energy storage
        self._add_curtail_ess_constraints(step_k)

        # Solar, wind, and import
        self._add_unit_link_pu(step_k)

        # Hydropower
        self._add_hydropower_constraints(step_k)

        # Update the model just in case we want to check model structure
        self.model.update()

    def build(
        self,
        step_k: int,
        init_conds: dict[str, dict],
    ) -> PowerSystemModel:
        """Build the model for the unit commitment problem."""
        self.model = gp.Model(self.inputs.model_id)
        self.add_variables(step_k=step_k)
        self.set_objfunc(step_k=step_k)
        self.add_constraints(step_k=step_k, init_conds=init_conds)
        return PowerSystemModel(self.model)

    def _update_variables(self, step_k: int) -> None:
        """
        Update the lower/upper bounds of time-dependent variables based on
        the current timestep, itereration, and capacity dataframes.

        This method dynamically adjusts the upper bounds (capacity limits)
        for various types of variables in the optimization problem,
        ensuring they reflect the available capacity at the given time step
        within the simulation horizon.

        Variables updated include:

        - Thermal units:
            - `pthermal`: Thermal power output
            - `vpower`: Thermal unit commitment status (on/off)
            - `vpowerbar`: Thermal unit start-up status
            - `pthermal_curtail`: Curtailed thermal power output
        - Energy storage:
            - `charge_state`: State of charge
        - Flow variables:
            - `flow`: Power flow on transmission lines (updates both lower and upper bounds)

        The upper bounds are retrieved from the corresponding capacity dataframes
        (e.g., `thermal_derated_capacity`, `hydro_capacity`, etc.) based on
        the current time step (`step_k`) and the simulation horizon.

        Args:
            step_k: The current time step within the simulation horizon.

        """

        def _update_var_ub(
            variables: gp.tupledict,
            capacity_df: pd.DataFrame,
        ) -> None:
            """Update the upper bounds of the variables based on the capacity dataframes"""
            for v in variables.values():
                unit_g, t = get_unit_hour_from_varnam(v.VarName)
                capacity_value = capacity_df.loc[t + (step_k - 1) * 24, unit_g]
                # Check if capacity_value is a Series or DataFrame and use .iloc[0] if necessary
                if isinstance(capacity_value, (pd.Series, pd.DataFrame)):
                    capacity_value = capacity_value.iloc[0]

                v.ub = capacity_value

        def _update_flow_bounds(
            flow_variables: gp.tupledict,
            capacity_df: pd.DataFrame,
        ) -> None:
            """Update the lower and upper bounds of the flow variables based on the capacity dataframes"""
            for flow_variable in flow_variables.values():
                edge, t = get_edge_hour_from_varname(flow_variable.VarName)
                line_capacity = capacity_df.loc[t + (step_k - 1) * 24, edge]
                # Update the lower and upper bounds, respectively
                flow_variable.lb = -line_capacity * self.inputs.line_capacity_factor
                flow_variable.ub = line_capacity * self.inputs.line_capacity_factor

        # Update variable bounds
        thermal_unit_vars = [
            self.pthermal,
            self.vpower,
            self.vpowerbar,
            self.pthermal_curtail,
        ]
        for var_dict in thermal_unit_vars:
            _update_var_ub(var_dict, self.inputs.thermal_derated_capacity)

        # Energy storage, renewables and import
        variable_capacity_pairs = [
            (self.charge_state, self.inputs.ess_derated_capacity),
        ]
        for var_dict, capacity_df in variable_capacity_pairs:
            _update_var_ub(var_dict, capacity_df)

        # Flow variables require updating both lower and upper bounds
        _update_flow_bounds(self.flow, self.inputs.line_capacity)

    def _update_objfunc(self, step_k: int) -> None:
        """Update the objective function with time-dependent terms."""
        rnw_import_storage_expr = self._build_rnw_import_storage_objfunc_terms(
            step_k=step_k
        )
        self.model.setObjective(
            self.thermal_fixed_expr
            + self.thermal_opex_expr
            + self.thermal_startup_expr
            + self.thermal_curtail_expr
            + self.load_shortfall_penalty_expr
            + self.spin_shortfall_penalty_expr
            + rnw_import_storage_expr
        )

    def _update_constraints(self, step_k, init_conds: dict) -> None:
        """Constraints to be updated include:

        *Thermal units*
        - c_link_uvw_init: initial_u is from the previous iteration
        - c_link_pu_upper: thermal_derated_capacity is a timeseries
        - c_min_down_init: initial_min_off is from the previous iteration
        - c_min_up_init: initial_min_on is from the previous iteration
        - c_ramp_down_init: initial vpower and u is from the previous iteration
        - c_ramp_up_init: initial vpower is from the previous iteration

        *System-wide*
        - c_angle_diff: susceptance is a timeseries
        - c_kirchhoff: Susceptance is a timeseries
        - c_flow_balance: Electricity demand is a timeseries
        - c_reserve_req: Spinning reserve requirement is based on the electricity demand

        *Upper bound of renewable units*
        - c_link_hydro_pu: Hydropower capacity is a timeseries
        - c_link_solar_pu: Solar capacity is a timeseries
        - c_link_wind_pu: Wind capacity is a timeseries
        - c_link_import_pu: Import capacity is a timeseries

        - c_hydro_limit_daily: Hydropower capacity is a timeseries

        *Curtailment*
        - c_hydro_curtail_ess: Hydropower capacity is a timeseries
        - c_daily_hydro_curtail_ess: Daily hydropower capacity is a timeseries
        - c_solar_curtail_ess: Solar capacity is a timeseries
        - c_wind_curtail_ess: Wind capacity is a timeseries
        - c_import_curtail_ess: Import capacity is a timeseries

        *Energy storage units*
        - c_unit_ess_balance_init: initial condition changes

        """
        self.model.remove(self.c_link_uvw_init)
        self.c_link_uvw_init = modeling.add_c_link_uvw_init(
            model=self.model,
            u=self.status,
            v=self.startup,
            w=self.shutdown,
            initial_u=init_conds["initial_u"],
            thermal_units=self.inputs.thermal_units,
        )

        self.model.remove(self.c_link_pu_upper)
        self.c_link_pu_upper = modeling.add_c_link_pu_upper(
            model=self.model,
            pbar=self.vpowerbar,
            u=self.status,
            timesteps=self.timesteps,
            step_k=step_k,
            thermal_units=self.inputs.thermal_units,
            thermal_min_capacity=self.inputs.thermal_min_capacity,
            thermal_derated_capacity=self.inputs.thermal_derated_capacity,
        )

        self.model.remove(self.c_thermal_curtail)
        self.c_thermal_curtail = modeling.add_c_thermal_curtail(
            model=self.model,
            pthermal=self.pthermal,
            pthermal_curtail=self.pthermal_curtail,
            pcharge=self.pcharge,
            timesteps=self.timesteps,
            step_k=step_k,
            thermal_derated_capacity=self.inputs.thermal_derated_capacity,
            thermal_must_take_units=self.inputs.thermal_must_take_units,
            ess_attached=self.inputs.ess_thermal_units,
        )

        self.model.remove(self.c_min_down_init)
        self.c_min_down_init = modeling.add_c_min_down_init(
            model=self.model,
            u=self.status,
            sim_horizon=self.inputs.sim_horizon,
            thermal_units=self.inputs.thermal_units,
            initial_min_off=init_conds["initial_min_off"],
        )

        self.model.remove(self.c_min_up_init)
        self.c_min_up_init = modeling.add_c_min_up_init(
            model=self.model,
            u=self.status,
            sim_horizon=self.inputs.sim_horizon,
            thermal_units=self.inputs.thermal_units,
            initial_min_on=init_conds["initial_min_on"],
        )

        self.model.remove(self.c_ramp_down_init)
        self.c_ramp_down_init = modeling.add_c_ramp_down_init(
            model=self.model,
            p=self.vpower,
            w=self.shutdown,
            thermal_units=self.inputs.thermal_units,
            initial_p=init_conds["initial_p"],
            initial_u=init_conds["initial_u"],
            thermal_min_capacity=self.inputs.thermal_min_capacity,
            RD=self.inputs.RD,
            SD=self.inputs.SD,
        )

        self.model.remove(self.c_ramp_up_init)
        self.c_ramp_up_init = modeling.add_c_ramp_up_init(
            model=self.model,
            pbar=self.vpowerbar,
            u=self.status,
            v=self.startup,
            thermal_units=self.inputs.thermal_units,
            initial_p=init_conds["initial_p"],
            thermal_min_capacity=self.inputs.thermal_min_capacity,
            RU=self.inputs.RU,
            SU=self.inputs.SU,
        )

        if self.inputs.dc_opf == "voltage_angle":
            self.model.remove(self.c_angle_diff)
            self.c_angle_diff = modeling.add_c_angle_diff(
                model=self.model,
                flow=self.flow,
                theta=self.theta,
                timesteps=self.timesteps,
                sim_horizon=self.inputs.sim_horizon,
                step_k=step_k,
                edges=self.inputs.edges,
                susceptance=self.inputs.susceptance,
            )

        elif self.inputs.dc_opf == "kirchhoff":
            # There might not be a cycle in the network
            if self.c_kirchhoff is not None:
                self.model.remove(self.c_kirchhoff)
                self.c_kirchhoff = modeling.add_c_kirchhoff(
                    model=self.model,
                    flow=self.flow,
                    timesteps=self.timesteps,
                    step_k=step_k,
                    edges=self.inputs.edges,
                    cycle_map=self.inputs.cycle_map,
                    susceptance=self.inputs.susceptance,
                )

        self.model.remove(self.c_flow_balance)
        self.c_flow_balance = modeling.add_c_flow_balance(
            model=self.model,
            pthermal=self.pthermal,
            phydro=self.phydro,
            psolar=self.psolar,
            pwind=self.pwind,
            pimp=self.pimp,
            pcharge=self.pcharge,
            pdis=self.pdischarge,
            pos_pmismatch=self.pos_pmismatch,
            neg_pmismatch=self.neg_pmismatch,
            flow=self.flow,
            timesteps=self.timesteps,
            step_k=step_k,
            thermal_units=self.inputs.thermal_units,
            hydro_units=self.inputs.hydro_units,
            solar_units=self.inputs.solar_units,
            wind_units=self.inputs.wind_units,
            import_units=self.inputs.import_units,
            nodes=self.inputs.nodes,
            node_edge=self.inputs.node_edge,
            node_generator=self.inputs.node_generator,
            charge_storage=self.inputs.ess_substation_units,
            discharge_storage=self.inputs.ess_attach_unit,
            demand_nodes=self.inputs.demand_nodes,
            demand=self.inputs.demand,
            line_loss_factor=self.inputs.line_loss_factor,
        )

        # Reserve requirement
        self.model.remove(self.c_reserve_req)
        if self.inputs.use_spin_var:
            self.c_reserve_req = modeling.add_c_reserve_req_1(
                model=self.model,
                spin=self.spin,
                spin_shortfall=self.spin_shortfall,
                timesteps=self.timesteps,
                step_k=step_k,
                thermal_units=self.inputs.thermal_units,
                spin_requirement=self.inputs.spin_requirement,
            )
        else:
            self.c_reserve_req = modeling.add_c_reserve_req_2(
                model=self.model,
                pbar=self.vpowerbar,
                u=self.status,
                spin_shortfall=self.spin_shortfall,
                timesteps=self.timesteps,
                step_k=step_k,
                thermal_units=self.inputs.thermal_units,
                thermal_min_capacity=self.inputs.thermal_min_capacity,
                demand_nodes=self.inputs.demand_nodes,
                demand=self.inputs.demand,
                spin_requirement=self.inputs.spin_requirement,
            )

        # Renewables curtailment
        self.model.remove(self.c_solar_curtail_ess)
        self.model.remove(self.c_wind_curtail_ess)
        self.model.remove(self.c_import_curtail_ess)
        self._add_curtail_ess_constraints(step_k)

        # Solar, wind, import link p and u.
        self.model.remove(self.c_link_solar_pu)
        self.model.remove(self.c_link_wind_pu)
        self.model.remove(self.c_link_import_pu)
        self._add_unit_link_pu(step_k)

        # Hydropower
        self._update_hydropower_constraints(step_k)

        # Flow balance
        self.model.remove(self.c_unit_ess_balance_init)
        self.c_unit_ess_balance_init = modeling.add_c_unit_ess_balance_init(
            model=self.model,
            pcharge=self.pcharge,
            pdischarge=self.pdischarge,
            charge_state=self.charge_state,
            units=self.inputs.storage_units,
            charge_state_init=init_conds["initial_charge_state"],
            charge_efficiency=self.inputs.ess_charge_efficiency,
            discharge_efficiency=self.inputs.ess_discharge_efficiency,
            self_discharge_rate=self.inputs.ess_self_discharge_rate,
        )

    def update(
        self,
        step_k: int,
        init_conds: dict[str, dict],
    ) -> gp.Model:
        """Update the model instead of creating a new one
        so we can perform warm start
        """
        self._update_variables(step_k=step_k)
        self._update_objfunc(step_k=step_k)
        self._update_constraints(step_k=step_k, init_conds=init_conds)
        self.model.update()

        return PowerSystemModel(self.model)

    def print_added_constraints(self):
        added_constrs = []
        not_added_constrs = []

        constraints_list = [
            "c_link_uvw_init",
            "c_link_uvw",
            "c_link_pthermal",
            "c_link_pu_lower",
            "c_link_pu_upper",
            "c_thermal_curtail",
            "c_min_down_init",
            "c_min_up_init",
            "c_min_down",
            "c_min_up",
            "c_peak_down_bound",
            "c_peak_up_bound",
            "c_ramp_down_init",
            "c_ramp_up_init",
            "c_ramp_down",
            "c_ramp_up",
            "c_ref_node",
            "c_angle_diff",
            "c_kirchhoff",
            "c_flow_balance",
            "c_link_spin",
            "c_link_ppbar",
            "c_reserve_req",
            "c_hydro_curtail_ess",
            "c_daily_hydro_curtail_ess",
            "c_solar_curtail_ess",
            "c_wind_curtail_ess",
            "c_import_curtail_ess",
            "c_link_hydro_pu",
            "c_link_daily_hydro_pu",
            "c_link_solar_pu",
            "c_link_wind_pu",
            "c_link_import_pu",
            "c_hydro_limit_daily",
            "c_link_ess_charge",
            "c_link_discharge",
            "c_link_ess_state",
            "c_unit_ess_balance_init",
            "c_unit_ess_balance",
        ]

        for attr_name in constraints_list:
            if getattr(self, attr_name):
                added_constrs.append(attr_name)
            else:
                not_added_constrs.append(attr_name)

        log_message = "\nAdded constraints:\n"
        log_message += "\n".join(added_constrs)
        log_message += "\n\nNot added constraints:\n"
        log_message += "\n".join(not_added_constrs)

        logger.warning(log_message)



""" output.py: Contains the OutputProcessor class, which processes the output from PowNet and provides methods to access the data.
"""

import pandas as pd

from .input import SystemInput
from pownet.data_utils import get_dates, get_fuel_mix_order


class OutputProcessor:
    def __init__(self) -> None:
        """Initialize the OutputProcessor object."""
        self.year: int = None
        self.fuelmap: dict = {}
        self.dates: pd.DataFrame = pd.DataFrame()

    def load(self, inputs: SystemInput) -> None:
        """Load the input data."""
        self.fuelmap = inputs.fuelmap
        self.year = inputs.year

        self.dates = get_dates(year=self.year)
        self.dates.index += 1

    def _get_power_variables(self, node_variables: pd.DataFrame) -> pd.DataFrame:
        power_vars = {
            "pthermal",
            "psolar",
            "pwind",
            "phydro",
            "pimp",
            "pos_pmismatch",
            "neg_pmismatch",
            "pdischarge",
            "pcharge",
        }
        power_variables = node_variables.loc[
            node_variables["vartype"].isin(power_vars)
        ].reset_index(drop=True)
        power_variables["fuel_type"] = power_variables.apply(
            lambda x: self.fuelmap.get(x["node"], None), axis=1
        )
        # These variables are missing in SystemInputs.fuelmap
        vartype_to_fuel_type = {
            "pimp": "import",
            "pos_pmismatch": "shortfall",
            "neg_pmismatch": "curtailment",
            "pdischarge": "discharging",
            "pcharge": "charging",
        }
        for vartype, fuel_type in vartype_to_fuel_type.items():
            power_variables.loc[power_variables["vartype"] == vartype, "fuel_type"] = (
                fuel_type
            )
        # Convert charging to negative values for plotting
        power_variables.loc[power_variables["vartype"] == "pcharge", "value"] *= -1
        return power_variables

    def get_hourly_generation(self, node_variables: pd.DataFrame) -> pd.DataFrame:
        power_variables = self._get_power_variables(node_variables)
        hourly_generation = (
            power_variables[["fuel_type", "value", "hour"]]
            .groupby(["fuel_type", "hour"])
            .sum()
        )
        hourly_generation = hourly_generation.reset_index()
        hourly_generation = hourly_generation.pivot(
            columns=["hour"], index=["fuel_type"]
        ).T.reset_index(drop=True)
        # PowNet indexing starts at 1
        hourly_generation.index += 1
        hourly_generation.index.name = "Hour"

        # Define the order of fuels for plotting. Baseload at the bottom,
        # renewables in the middle, then peaker plants, and shortfall
        fuel_mix_order = get_fuel_mix_order()
        fuel_mix_order = [
            fuel for fuel in fuel_mix_order if fuel in hourly_generation.columns
        ]
        hourly_generation = hourly_generation[fuel_mix_order]
        return hourly_generation

    def get_daily_generation(self, node_variables: pd.DataFrame) -> pd.DataFrame:
        hourly_generation = self.get_hourly_generation(node_variables)
        # Sum across 24 hours to get the daily dispatch.
        daily_generation = hourly_generation.groupby(
            (hourly_generation.index - 1) // 24
        ).sum()
        daily_generation.index += 1
        daily_generation.index.name = "Day"
        return daily_generation

    def get_monthly_generation(self, node_variables: pd.DataFrame) -> pd.DataFrame:
        monthly_generation = self.get_hourly_generation(node_variables)
        monthly_generation["month"] = self.dates["date"].dt.to_period("M")
        monthly_generation = monthly_generation.groupby("month").sum()
        monthly_generation.index = monthly_generation.index.strftime("%b")
        monthly_generation.index.name = "Month"
        return monthly_generation

    def get_hourly_demand(self, input_demand: pd.DataFrame) -> pd.DataFrame:
        # Demand is an input to the simulation
        hourly_demand = input_demand.sum(axis=1).to_frame()
        hourly_demand.columns = ["demand"]
        hourly_demand.index.name = "Hour"
        return hourly_demand

    def get_daily_demand(self, input_demand: pd.DataFrame) -> pd.Series:
        hourly_demand = self.get_hourly_demand(input_demand)
        # Need -1 because the index starts with 1 and we want to group by 24 hours
        daily_demand = hourly_demand.groupby((hourly_demand.index - 1) // 24).sum()
        daily_demand.index += 1
        return daily_demand

    def get_monthly_demand(self, input_demand: pd.DataFrame) -> pd.Series:
        monthly_demand = self.get_hourly_demand(input_demand)
        monthly_demand["month"] = self.dates["date"].dt.to_period("M")
        monthly_demand = monthly_demand.groupby("month").sum()
        monthly_demand.index = monthly_demand.index.strftime("%b")
        monthly_demand.index.name = "Month"
        return monthly_demand

    def get_thermal_unit_hourly_status(
        self, node_variables: pd.DataFrame
    ) -> pd.DataFrame:
        """The hourly status of thermal units for each hour over the simulation period."""
        thermal_unit_hourly_status = node_variables[
            node_variables["vartype"] == "status"
        ].reset_index(drop=True)
        thermal_unit_hourly_status["timestep"] = thermal_unit_hourly_status["hour"] % 24
        return thermal_unit_hourly_status

    def get_thermal_unit_mean_hourly_status(
        self, node_variables: pd.DataFrame
    ) -> pd.DataFrame:
        """The hourly status of thermal units for each hour over the simulation period."""
        thermal_unit_hourly_status = node_variables[
            node_variables["vartype"] == "status"
        ].reset_index(drop=True)
        thermal_unit_hourly_status["timestep"] = thermal_unit_hourly_status["hour"] % 24
        return thermal_unit_hourly_status.pivot_table(
            columns="node", index="timestep", values="value", aggfunc="mean"
        )

    def get_thermal_unit_daily_duration(
        self, node_variables: pd.DataFrame
    ) -> pd.DataFrame:
        """Return the daily online duration of each thermal unit. Rows are days and columns are units."""
        status_variables = node_variables[node_variables["vartype"] == "status"].copy()
        status_variables["day"] = (status_variables["hour"] - 1) // 24 + 1
        return status_variables.pivot_table(
            columns="node", index="day", values="value", aggfunc="sum"
        )

    def get_thermal_unit_startup_frequency(
        self, node_variables: pd.DataFrame
    ) -> pd.DataFrame:
        """Return the frequency of startups for each thermal unit over the whole simulation period."""
        startup_vars = node_variables[node_variables["vartype"] == "startup"].copy()
        startup_vars["day"] = (startup_vars["hour"] - 1) // 24 + 1
        return startup_vars.pivot_table(
            columns="node", index="day", values="value", aggfunc="sum"
        )

    def get_thermal_unit_total_duration_and_frequency(
        self, node_variables: pd.DataFrame
    ) -> pd.DataFrame:
        """Return data for histogram of frequency of startups and duration of committed hours in a year."""
        cols = ["startup", "status"]
        data = node_variables[node_variables["vartype"].isin(cols)]
        # Sum the number of startups and committed hours for each thermal unit
        return (
            data.groupby(["node", "vartype"]).sum().reset_index().drop(columns=["hour"])
        )

    def get_thermal_unit_daily_dispatch(
        self, node_variables: pd.DataFrame
    ) -> pd.DataFrame:
        pthermal = node_variables[node_variables["vartype"] == "pthermal"].copy()
        pthermal["day"] = (pthermal["hour"] - 1) // 24 + 1
        return pthermal.pivot_table(
            columns="node", index="day", values="value", aggfunc="sum"
        )

    def get_thermal_unit_hourly_dispatch(
        self, node_variables: pd.DataFrame
    ) -> pd.DataFrame:
        return node_variables[node_variables["vartype"] == "pthermal"].copy()

    def get_nondispatch_hourly_capacity_factor(
        self,
        type: str,
        node_variables: pd.DataFrame,
        unit_capacities: dict[str, float],
        energy_storage_attach: dict[str, str],
    ) -> pd.DataFrame:
        """Return the capacity factor which is a function of generation and storage charging."""
        type_map = {
            "hydropower": "phydro",
            "solar": "psolar",
            "wind": "pwind",
            "import": "pimp",
        }
        # Power output variables
        generation = (
            node_variables[node_variables["vartype"] == type_map[type]]
            .copy()
            .drop(columns=["vartype"])
        )
        units = generation["node"].unique()
        generation = generation.set_index(["node", "hour"])

        # Process charging variables
        charging = node_variables[node_variables["vartype"] == "pcharge"].copy()
        if not charging.empty:
            charging["unit"] = charging.apply(
                lambda x: energy_storage_attach.get(x["node"], None), axis=1
            )
            charging = charging[charging["unit"].isin(units)]
            charging = charging.drop(columns=["vartype", "node"]).set_index(
                ["unit", "hour"]
            )
        # Capacity factor is the sum of generation and charging divided by the unit capacity
        output = generation + charging
        output = output.reset_index()
        output["capacity_factor"] = output.apply(
            lambda x: x["value"] / unit_capacities[x["node"]], axis=1
        )
        return output.pivot(columns="node", index="hour", values="capacity_factor")

    def get_energy_storage_hourly_charge(
        self, node_variables: pd.DataFrame
    ) -> pd.DataFrame:
        hourly_charge = (
            node_variables[node_variables["vartype"] == "pcharge"]
            .copy()
            .drop(columns=["vartype"])
        )
        return hourly_charge.pivot(columns="node", index="hour", values="value")

    def get_energy_storage_hourly_discharge(
        self, node_variables: pd.DataFrame
    ) -> pd.DataFrame:
        hourly_discharge = (
            node_variables[node_variables["vartype"] == "pdischarge"]
            .copy()
            .drop(columns=["vartype"])
        )
        return hourly_discharge.pivot(columns="node", index="hour", values="value")

    def _get_hourly_charge_state_fraction(
        self, node_variables: pd.DataFrame, max_storage: dict[str, float]
    ) -> pd.DataFrame:
        hourly_state = (
            node_variables[node_variables["vartype"] == "charge_state"]
            .copy()
            .drop(columns=["vartype"])
        )
        hourly_state["storage_fraction"] = hourly_state.apply(
            lambda x: x["value"] / max_storage[x["node"]], axis=1
        )
        return hourly_state

    def get_energy_storage_hourly_state(
        self,
        node_variables: pd.DataFrame,
        max_storage: dict[str, float],
    ) -> pd.DataFrame:
        hourly_storage_state = self._get_hourly_charge_state_fraction(
            node_variables, max_storage
        )
        return hourly_storage_state.pivot(
            columns="node", index="hour", values="storage_fraction"
        )

    def get_energy_storage_daily_state(
        self,
        node_variables: pd.DataFrame,
        max_storage: dict[str, float],
    ) -> pd.DataFrame:
        hourly_storage_state = self._get_hourly_charge_state_fraction(
            node_variables, max_storage
        )
        hourly_storage_state["day"] = (hourly_storage_state["hour"] - 1) // 24 + 1
        return hourly_storage_state.pivot_table(
            columns="node", index="day", values="storage_fraction", aggfunc="last"
        )

    def get_import_values(self, node_variables: pd.DataFrame) -> pd.DataFrame:
        """Return the import values for each timestep. Columns are generators.
        Index is the hour in the simulation year"""
        power_variables = self._get_power_variables(node_variables)
        import_values = power_variables[power_variables["vartype"] == "pimp"]
        import_values = import_values.pivot(
            columns="node", index="hour", values="value"
        )
        return import_values

    def get_co2_emission(self, co2_map: dict[str:float]) -> pd.DataFrame:
        """Return the CO2 emissions for timestep.
        From Chowdhury, Dang, Nguyen, Koh, & Galelli. (2021).

        coal: 1.04 Mton/MWh
        gas:  0.47 Mton/MWh
        oil : 0.73 Mton/MWh
        solid_waste: 0.170 Mton/MWh

        From https://www.eia.gov/environment/emissions/co2_vol_mass.php:
        solid_waste: 49.89 kg/MMBtu
              = 49.89 kg/MMBtu * 3.412 MMBtu/MWh * 1 Mton/1000 kg = 0.170
        """
        if co2_map is None:
            co2_map = {
                "coal": 1.04,
                "gas": 0.47,
                "oil": 0.73,
                "import": 0.0,
                "shortfall": 0.0,
                "curtailment": 0.0,
                "biomass": 0.0,
                "solid_waste": 0.170,
                "slack": 0.0,
            }

        df = self.get_hourly_thermal_dispatch()
        co2_emissions = pd.DataFrame()
        for fuel in df.columns:
            co2_emissions[fuel] = df[fuel] * co2_map[fuel]
        return co2_emissions

    def get_max_line_usage(
        self,
        flow_variables: pd.DataFrame,
        line_locations: pd.DataFrame,
        rated_line_capacities: dict[tuple[str, str], int],
    ) -> pd.DataFrame:

        # Prevent unintentional modification to the original dataframe
        flow_variables = flow_variables.copy()

        flow_variables = flow_variables.rename(
            columns={"node_a": "source", "node_b": "sink"}
        ).drop("hour", axis=1)
        # Flow can be negative due to flow being directional
        flow_variables["value"] = flow_variables["value"].abs()

        # Find the max_value for each line segment across the whole time horizon
        flow_variables["max_value"] = flow_variables.groupby(["source", "sink"])[
            "value"
        ].transform("max")
        # Drop duplicates because we are only interested in the maximum flow
        # over the whole simulation
        flow_variables = flow_variables.drop_duplicates(subset=["source", "sink"])

        # Max utilization rate
        flow_variables["max_line_usage"] = flow_variables.apply(
            lambda row: row["max_value"]
            / rated_line_capacities[(row["source"], row["sink"])],
            axis=1,
        ).round(4)

        flow_variables = flow_variables[["source", "sink", "max_line_usage"]].set_index(
            ["source", "sink"]
        )
        flow_variables = flow_variables.merge(
            line_locations, how="left", left_index=True, right_index=True
        )
        # Append rated capacities
        flow_variables["rated_capacity"] = [
            rated_line_capacities[idx] for idx in flow_variables.index
        ]
        return flow_variables

    def get_fuel_mix(self, hourly_generation: pd.DataFrame) -> pd.DataFrame:
        """Return the fuel mix (%) for the whole simulation period."""
        return (hourly_generation.sum() / hourly_generation.sum().sum()).round(4) * 100

    def get_gen_by_fuel(self, hourly_generation: pd.DataFrame) -> pd.DataFrame:
        """Return the total generation for the whole simulation period."""
        return hourly_generation.sum().round(0)

    def get_curtailed_power(self):
        """Return the statistics of curtailed energy by hydro, solar, wind, and import."""
        pass

    def get_contract_hourly_generation(
        self, node_variables: pd.DataFrame, unit_contract: dict[str, str]
    ) -> pd.DataFrame:
        power_variables = self._get_power_variables(node_variables)
        vartypes = ["pthermal", "phydro", "psolar", "pwind", "pimp", "pdischarge"]
        power_variables = power_variables[power_variables["vartype"].isin(vartypes)]
        power_variables["contract"] = power_variables.apply(
            lambda x: unit_contract.get(x["node"], None), axis=1
        )
        return power_variables[["contract", "hour", "value"]]

    def get_contract_generation(
        self, node_variables: pd.DataFrame, unit_contract: dict[str, str]
    ) -> pd.DataFrame:
        hourly_generation = self.get_contract_hourly_generation(
            node_variables=node_variables, unit_contract=unit_contract
        )
        return hourly_generation[["contract", "value"]].groupby("contract").sum()

    def get_contract_cost(
        self,
        node_variables: pd.DataFrame,
        unit_contract: dict[str, str],
        contract_costs: dict[str, float],
    ) -> pd.DataFrame:
        hourly_generation = self.get_contract_hourly_generation(
            node_variables=node_variables, unit_contract=unit_contract
        )
        hourly_generation["cost"] = hourly_generation.apply(
            lambda x: x["value"] * contract_costs[(x.contract, x.hour)], axis=1
        )
        return (
            hourly_generation.drop(columns=["hour", "value"]).groupby("contract").sum()
        )



""" variable.py: Contains functions for adding variables to the optimization model.
"""

import math
import gurobipy as gp
from gurobipy import GRB
import pandas as pd


"""
Functions for adding thermal variables
"""


def add_var_with_variable_ub(
    model: gp.Model,
    varname: str,
    timesteps: range,
    step_k: int,
    units: list,
    capacity_df: pd.DataFrame,
) -> gp.tupledict:
    def get_capacity_value(t, unit_g):
        value = capacity_df.loc[t + (step_k - 1) * 24, unit_g]
        if isinstance(value, pd.Series):
            return value.iloc[0]
        return value

    return model.addVars(
        units,
        timesteps,
        lb=0,
        ub={
            (unit_g, t): get_capacity_value(t, unit_g)
            for t in timesteps
            for unit_g in units
        },
        vtype=GRB.CONTINUOUS,
        name=varname,
    )


def add_var_flow(
    model: gp.Model,
    timesteps: range,
    step_k: int,
    edges: list,
    line_capacity_factor: float,
    line_capacity: pd.DataFrame,
) -> gp.tupledict:
    """The flow variable f(a,b,t) represents the flow in the
    line segment (a,b) at hour t in MW/hr). If the flow is positive,
    then energy flows from a to b. Unit: MW/hr.

    Args:
        model (gp.Model): The optimization model.
        timesteps (range): The range of timesteps.
        step_k (int): The step index.
        sim_horizon (int): The simulation horizon.
        edges (list): The list of edges.
        line_capacity_factor (float): The line capacity factor.
        line_capacity (pd.DataFrame): The dataframe of line capacities.

    Returns:
        gp.tupledict: The flow variable representing the flow in each line segment
    """
    return model.addVars(
        edges,
        timesteps,
        lb={
            (source, sink, t): -1
            * line_capacity_factor
            * line_capacity.loc[t + (step_k - 1) * 24, (source, sink)]
            for t in timesteps
            for source, sink in edges
        },
        ub={
            (source, sink, t): line_capacity_factor
            * line_capacity.loc[t + (step_k - 1) * 24, (source, sink)]
            for t in timesteps
            for source, sink in edges
        },
        vtype=GRB.CONTINUOUS,
        name="flow",
    )


def add_var_voltage_angle(
    model: gp.Model, timesteps: range, nodes: list
) -> gp.tupledict:
    """The voltage angle of a node. Unit: radians.
    This refers to the angular difference between the voltage phasor
    at a particular bus and the reference bus (usually the slack bus).

    NOTE: The voltage angle is bounded between -pi and pi.
    This choice simulates the theoretical limit of the voltage angle.
    Since the assumption of DC-OPF is that the voltage angle is small,
    some studies use the bound of -pi/2 and pi/2.

    Args:
        model (gp.Model): The optimization model.
        timesteps (range): The range of timesteps.
        nodes (list): The list of nodes.

    Returns:
        gp.tupledict: The voltage angle of each node.

    """
    return model.addVars(
        nodes,
        timesteps,
        lb=-math.pi,
        ub=math.pi,
        vtype=GRB.CONTINUOUS,
        name="theta",
    )



""" record.py: This module contains the SystemRecord class, which processes stores the modeling outputs from each iteration.
TODO: self.current_hydro, self.current_import for model coupling
"""

from .input import SystemInput

import pandas as pd

from pownet.data_utils import (
    parse_node_variables,
    parse_flow_variables,
    parse_syswide_variables,
    parse_lmp,
    calc_remaining_on_duration,
    calc_remaining_off_duration,
    write_df,
)


class SystemRecord:
    """This class stores modeling outputs from each iteration. The results are stored
    in three dataframes: node_vars, flow_vars, and syswide_vars. The initial conditions
    are also stored in the class.
    """

    def __init__(
        self,
        system_input: SystemInput,
        batch_mode: bool = True,
        keep_record_each_step: bool = False,
    ) -> None:
        """Initialize the SystemRecord object.
        Args:
            system_input (SystemInput): The input object containing the simulation parameters.
        """

        self.inputs: SystemInput = system_input
        self.batch_mode: bool = batch_mode
        self.keep_record_each_step: bool = keep_record_each_step

        if not batch_mode and not keep_record_each_step:
            print(
                '\nWarning: No data will be stored because both "batch_mode" and "keep_record_each_step" are False.'
            )

        # Format of variable name: var(node, t)
        self.node_vars: pd.DataFrame = pd.DataFrame()
        # Format of variable name: flow(node_a, node_b, t)
        self.flow_vars: pd.DataFrame = pd.DataFrame()
        # Format of variable name: var(t)
        self.syswide_vars: pd.DataFrame = pd.DataFrame()

        # Locational marginal prices (LMP)
        self.lmp_df: pd.DataFrame = pd.DataFrame()

        self.objvals: list = []
        self.runtimes: list = []

        # These are vpower, unit status, unit switching, etc.
        self.current_p: dict[str] = {}
        self.current_u: dict[str] = {}
        self.current_v: dict[str] = {}
        self.current_w: dict[str] = {}
        self.current_min_on: dict[str] = {}
        self.current_min_off: dict[str] = {}
        self.current_charge_state: dict[str] = {}

    def keep(
        self,
        runtime: float,
        objval: float,
        solution: pd.DataFrame,
        step_k: int,
        lmp: dict[str, float] = None,
    ) -> None:
        """Keep the simulation results at the current simulation period step_k.

        Args:
            runtime (float): The runtime of the model.
            objval (float): The objective value of the model.
            solution (pd.DataFrame): The solution dataframe from the model.
            step_k (int): The current simulation period.
            lmp (dict[str, float], optional): The locational marginal prices. Defaults to None.

        Returns:
            None
        """

        def _extract_vartype_data(df: pd.DataFrame, vartype: str) -> dict[str, float]:
            """Extracts data for a specific 'vartype' from the DataFrame and converts it to a dictionary."""
            return (
                df[
                    (df["vartype"] == vartype) & (df["timestep"] == 24)
                ]  # Only considers values of the last hour
                .drop("vartype", axis=1)
                .set_index(["node"])  # Assume generator names do not repeat
                .to_dict()["value"]
            )

        self.runtimes.append(runtime)
        self.objvals.append(objval)

        # Create a col of variable types for filtering
        pat_vartype = r"(\w+)\["
        solution[["vartype"]] = solution["varname"].str.extract(
            pat_vartype, expand=True
        )

        node_vars = parse_node_variables(solution, self.inputs.sim_horizon, step_k)
        # Only keep 24-hours as we are doing rolling horizon
        node_vars = node_vars[node_vars["timestep"] <= 24]
        ##################
        # Initial conditions: vpower (p), commitment (u),
        # startup (v), shutdown (w), and storage's charge_state
        ##################
        self.current_p = _extract_vartype_data(node_vars, "vpower")
        self.current_u = _extract_vartype_data(node_vars, "status")
        self.current_v = _extract_vartype_data(node_vars, "startup")
        self.current_w = _extract_vartype_data(node_vars, "shutdown")
        self.current_charge_state = _extract_vartype_data(node_vars, "charge_state")

        # Need to calculate the minimum time on/off
        self.current_min_on = calc_remaining_on_duration(
            node_vars,
            sim_horizon=24,
            thermal_units=self.inputs.thermal_units,
            TU=self.inputs.TU,
        )
        self.current_min_off = calc_remaining_off_duration(
            node_vars,
            sim_horizon=24,
            thermal_units=self.inputs.thermal_units,
            TD=self.inputs.TD,
        )

        ##################
        # Locational Marginal Prices (LMP)
        ##################
        if lmp is not None:
            self.lmp_df = pd.concat(
                [
                    self.lmp_df,
                    parse_lmp(lmp, self.inputs.sim_horizon, step_k).drop(
                        "timestep", axis=1
                    ),
                ],
                axis=0,
            )

        ##################
        # Append results to the existing dataframes in batch mode
        # otherwise, write to disk at each step_k if specified
        ##################
        # Keep outputs from the first 24-hr under rolling horizon for day-ahead planning
        node_vars = node_vars[node_vars["timestep"] <= 24]
        node_vars = node_vars.drop(["varname", "timestep"], axis=1)

        flow_vars = parse_flow_variables(
            solution=solution, sim_horizon=self.inputs.sim_horizon, step_k=step_k
        )
        flow_vars = flow_vars[flow_vars["timestep"] <= 24]
        flow_vars = flow_vars.drop("timestep", axis=1)

        syswide_vars = parse_syswide_variables(
            solution=solution, sim_horizon=self.inputs.sim_horizon, step_k=step_k
        )
        syswide_vars = syswide_vars[syswide_vars["timestep"] <= 24]
        syswide_vars = syswide_vars.drop("timestep", axis=1)

        if self.batch_mode:
            self.node_vars = pd.concat([self.node_vars, node_vars], axis=0)
            self.flow_vars = pd.concat([self.flow_vars, flow_vars], axis=0)
            self.syswide_vars = pd.concat([self.syswide_vars, syswide_vars], axis=0)

        if self.keep_record_each_step:
            output_folder = f"{self.inputs.model_id}_outputs"
            data_to_write = [
                (node_vars, f"node_variables_{step_k}"),
                (flow_vars, f"flow_variables_{step_k}"),
                (syswide_vars, f"system_variables_{step_k}"),
                (
                    pd.DataFrame({"objval": [objval], "runtime": [runtime]}),
                    f"model_stats_{step_k}",
                ),
            ]
            for df, output_name in data_to_write:
                write_df(
                    df,
                    output_folder=output_folder,
                    output_name=output_name,
                    model_id=self.inputs.model_id,  # Use model_name as the identifier
                )

    def get_init_conds(self) -> dict[str, dict]:
        """Return the initial conditions for the simulation."""
        return {
            "initial_p": self.current_p,
            "initial_u": self.current_u,
            "initial_v": self.current_v,
            "initial_w": self.current_w,
            "initial_min_on": self.current_min_on,
            "initial_min_off": self.current_min_off,
            "initial_charge_state": self.current_charge_state,
        }

    def get_node_variables(self) -> pd.DataFrame:
        """Return node-specific variables. These variables include
        dispatch, unit status, unit switching, etc.
        """
        return self.node_vars

    def get_flow_variables(self) -> pd.DataFrame:
        """Return the flow variables."""
        return self.flow_vars

    def get_systemwide_variables(self) -> pd.DataFrame:
        """Return the system variables. We currently only have the
        system-wide spinning reserve shortfall.
        """
        return self.syswide_vars

    def get_runtimes(self) -> list[float]:
        """Return the model runtime from each iteration."""
        return self.runtimes

    def get_objvals(self) -> list[float]:
        """Return the objective values from each iteration."""
        return self.objvals

    def get_lmp(self) -> pd.DataFrame:
        """Return the locational marginal prices (LMP) data."""
        return self.lmp_df.pivot_table(
            index="hour", columns="node", values="value", aggfunc="first"
        )

    def write_simulation_results(self, output_folder: str) -> None:
        """
        Write CSV files containing modeling results to the output directory.

        Args:
            output_folder (str): The directory where the output files will be saved.

        Returns:
            None
        """
        if not self.batch_mode:
            print("No data to write because batch_mode is False.")
            return

        data_to_write = [
            (self.node_vars, "node_variables"),
            (self.flow_vars, "flow_variables"),
            (self.syswide_vars, "system_variables"),
            (
                pd.DataFrame({"objval": self.objvals, "runtime": self.runtimes}),
                "model_stats",
            ),
        ]
        for df, output_name in data_to_write:
            write_df(
                df,
                output_folder=output_folder,
                output_name=output_name,
                model_id=self.inputs.model_id,  # Use model_name as the identifier
            )

        # Objective values and runtimes
        write_df(
            pd.DataFrame({"objval": self.objvals, "runtime": self.runtimes}),
            output_folder=output_folder,
            output_name="model_stats",
            model_id=self.inputs.model_id,
        )

        # LMP data if it exists
        if not self.lmp_df.empty:
            write_df(
                self.lmp_df,
                output_folder=output_folder,
                output_name="lmp",
                model_id=self.inputs.model_id,
            )



""" test_reservoir_functions.py
"""

import unittest

import pandas as pd

from pownet.reservoir.reservoir_functions import (
    adjust_hydropeaking,
    calc_target_level,
    calc_target_storage,
    calc_daily_hydropower,
)


class TestAdjustHydropeaking(unittest.TestCase):

    def test_no_change_needed(self):
        """Test when the release is already within bounds and no adjustment is needed."""
        result = adjust_hydropeaking(
            release=50, release_t0=45, max_release=100, min_release=20
        )
        self.assertEqual(result, 50)  # No change expected

    def test_increase_within_limit(self):
        """Test when the release needs to increase, but the change is within the hydropeaking limit."""
        result = adjust_hydropeaking(
            release=60,
            release_t0=40,
            max_release=100,
            min_release=20,
            hydropeak_factor=0.2,
        )
        self.assertEqual(result, 60)  # Increase allowed

    def test_increase_exceeds_limit(self):
        """Test when the release needs to increase, but the change exceeds the hydropeaking limit."""
        result = adjust_hydropeaking(
            release=80,
            release_t0=30,
            max_release=100,
            min_release=20,
            hydropeak_factor=0.1,
        )
        self.assertEqual(result, 40)  # Increase limited by hydropeaking factor

    def test_decrease_below_minimum(self):
        """Test when the release needs to decrease, but the adjusted value would be below the minimum."""
        result = adjust_hydropeaking(
            release=15,
            release_t0=40,
            max_release=100,
            min_release=20,
            hydropeak_factor=0.3,
        )
        self.assertEqual(result, 20)  # Decrease limited by minimum release

    def test_enforce_minimum(self):
        """Test when the adjusted release needs to be enforced to the minimum."""
        release = 10
        release_t0 = 30
        max_release = 100
        min_release = 20
        hydropeak_factor = 0.15

        expected_release = 20  # Enforced to minimum
        actual_release = adjust_hydropeaking(
            release, release_t0, max_release, min_release, hydropeak_factor
        )
        self.assertEqual(actual_release, expected_release)

    def test_enforce_maximum(self):
        """Test when the adjusted release needs to be enforced to the maximum."""
        release = 120
        release_t0 = 80
        max_release = 100
        min_release = 20
        hydropeak_factor = 0.20

        expected_release = 100  # Enforced to maximum
        actual_release = adjust_hydropeaking(
            release, release_t0, max_release, min_release, hydropeak_factor
        )
        self.assertEqual(actual_release, expected_release)


class TestCalcTargetLevel(unittest.TestCase):

    def test_basic_interpolation(self):
        """Test basic linear interpolation between min_day and max_day."""
        min_day = 100
        max_day = 200
        min_level = 5.0
        max_level = 10.0

        target_levels = calc_target_level(min_day, max_day, min_level, max_level)

        # Check specific values
        self.assertAlmostEqual(
            target_levels[min_day], min_level
        )  # Min level on min_day
        self.assertAlmostEqual(
            target_levels[max_day], max_level
        )  # Max level on max_day
        self.assertAlmostEqual(target_levels[(min_day + max_day) // 2], 7.5)  # Midpoint

    def test_interpolation_after_max(self):
        """Test interpolation for days after max_day."""
        min_day = 50
        max_day = 150
        min_level = 0.0
        max_level = 100.0

        target_levels = calc_target_level(min_day, max_day, min_level, max_level)

        # Check values after max_day
        self.assertAlmostEqual(target_levels[max_day + 1], 264 / 265 * 100)
        self.assertAlmostEqual(target_levels[365], 50 / 265 * 100)

    def test_interpolation_before_min(self):
        """Test interpolation for days before min_day."""
        min_day = 100
        max_day = 300
        min_level = 20.0
        max_level = 80.0

        target_levels = calc_target_level(min_day, max_day, min_level, max_level)

        # Check values before min_day
        self.assertAlmostEqual(target_levels[min_day - 1], 60 / 165 + 20)
        self.assertAlmostEqual(target_levels[1], (99 / 165 * 60) + 20)


class TestCalcTargetStorage(unittest.TestCase):

    def test_basic_calculation(self):
        """Test basic storage calculation with varying target levels."""
        target_level = pd.Series(
            [25.0, 50.0, 75.0], index=range(1, 4)
        )  # Example target levels
        min_level = 0.0
        max_level = 100.0
        max_storage = 1000.0

        expected_storage = pd.Series([250.0, 500.0, 750.0], index=range(1, 4))
        actual_storage = calc_target_storage(
            target_level, min_level, max_level, max_storage
        )

        pd.testing.assert_series_equal(actual_storage, expected_storage)


class TestCalcDailyHydropower(unittest.TestCase):

    def test_basic_calculation(self):
        def _calc_hydropower(release, mid_level, max_head, max_level):
            return (
                min(
                    0.85
                    * 998
                    * 9.81
                    * (max_head - (max_level - mid_level))
                    * (release / (24 * 3600))
                    / 1e6,
                    10.0,
                )
                * 24
            )

        """Test basic hydropower calculation with typical values."""
        release = pd.Series(
            [100000.0, 200000.0, 300000.0],
            index=range(1, 4),
        )  # Example release values (m3/day)
        mid_level = pd.Series(
            [90.0, 95.0, 100.0], index=range(1, 4)
        )  # Example mid-level values (m)
        max_generation = 10.0  # MW
        turbine_factor = 0.85
        max_head = 100.0  # m
        max_level = 105.0  # m

        # Calculate expected hydropower based on the formula
        expected_hydropower = pd.Series(
            [
                _calc_hydropower(release[1], mid_level[1], max_head, max_level),
                _calc_hydropower(release[2], mid_level[2], max_head, max_level),
                _calc_hydropower(release[3], mid_level[3], max_head, max_level),
            ],
            index=range(1, 4),
        )

        actual_hydropower = calc_daily_hydropower(
            release, mid_level, max_generation, turbine_factor, max_head, max_level
        )

        pd.testing.assert_series_equal(actual_hydropower, expected_hydropower)

    def test_max_generation_limit(self):
        """Test if hydropower is limited by max_generation."""
        release = pd.Series(
            [1000000.0],
        )  # Large release to potentially exceed max_generation
        mid_level = pd.Series([100.0])  # Example mid-level value (m)
        max_generation = 5.0  # MW
        turbine_factor = 0.9
        max_head = 50.0  # m
        max_level = 100.0  # m

        expected_hydropower = pd.Series(
            [max_generation * 24]
        )  # Hydropower limited to max_generation
        actual_hydropower = calc_daily_hydropower(
            release, mid_level, max_generation, turbine_factor, max_head, max_level
        )

        pd.testing.assert_series_equal(actual_hydropower, expected_hydropower)

    # def test_no_level_change(self):
    #     """Test when no level change with output at max head"""
    #     release = pd.Series([100000.0])
    #     mid_level = pd.Series(
    #         [105.0]
    #     )  # mid_level equals max_level, resulting in zero head
    #     max_generation = 10.0
    #     turbine_factor = 0.8
    #     max_head = 20.0
    #     max_level = 105.0

    #     expected_hydropower = pd.Series([0.0])
    #     actual_hydropower = calc_daily_hydropower(
    #         release, mid_level, max_generation, turbine_factor, max_head, max_level
    #     )

    #     pd.testing.assert_series_equal(actual_hydropower, expected_hydropower)


if __name__ == "__main__":
    unittest.main()



""" This is the model module.
"""

from .model import PowerSystemModel
from .variable import (
    add_var_with_variable_ub,
    add_var_flow,
    add_var_voltage_angle,
)

from .objfunc import (
    get_thermal_fixed_coeff,
    get_thermal_opex_coeff,
    get_thermal_startup_coeff,
    get_marginal_cost_coeff,
)

from .constraint import (
    add_c_link_uvw_init,
    add_c_link_uvw,
    add_c_link_pthermal,
    add_c_link_spin,
    add_c_link_ppbar,
    add_c_link_pu_lower,
    add_c_link_pu_upper,
    add_c_thermal_curtail,
    add_c_min_down_init,
    add_c_min_up_init,
    add_c_min_down,
    add_c_min_up,
    add_c_peak_down_bound,
    add_c_peak_up_bound,
    add_c_ramp_down_init,
    add_c_ramp_down,
    add_c_ramp_up_init,
    add_c_ramp_up,
    add_c_link_shortfall,
    add_c_ref_node,
    add_c_angle_diff,
    add_c_kirchhoff,
    add_c_flow_balance,
    add_c_reserve_req_1,
    add_c_reserve_req_2,
    add_c_hydro_limit_daily,
    add_c_link_unit_pu,
    add_c_link_unit_pu_constant,
    add_c_unit_curtail_ess,
    add_c_unit_curtail_ess_daily,
    add_c_link_ess_charge,
    add_c_link_ess_discharge,
    add_c_link_ess_state,
    add_c_unit_ess_balance_init,
    add_c_unit_ess_balance,
)
